#!/usr/bin/python

# ---------------------------------------------------------------------------
# Align the method parameters in HTML API documentation. This works with
# Java documentation generated by javadoc, and PHP documentation generated
# by ApiGen.
#
# This was orginally intended as a fix for Javadocs generated by Java 7,
# which doesn't properly align method parameters, making the docs look messy
# and difficult to read, but it now goes beyond that.
#
# For Java documentation, it will align the method parameter types and names
# in vertical columns, and also re-indent any "throws" declarations to make
# the docs easier to read. Any "throws" clause is indented by 4 spaces, and
# the thrown exceptions are all aligned after that, so that the exceptions
# are more easily distinguishable from the method parameters.
#
# For PHP documentation generated by ApiGen, it will align the method
# parameter types and names in vertical columns, rather than using a single
# line that wraps.
# ---------------------------------------------------------------------------
#
# ---------------------------------------------------------------------------
# Usage: align-method-params.py [doc_dir]
#
# This recursively traverses the directory hierarchy processing each HTML
# file found (modifying files in place). If the directory is omitted, it
# defaults to the current directory.
# ---------------------------------------------------------------------------

import os
import re
import sys

def process_file(filename):
    sys.stdout.write('  Processing file "%s" ... ' % os.path.basename(filename))

    with open(filename, 'r') as f:
        docs = f.read()

    # Test if this is PHP documentation generated by ApiGen. Otherwise, it is
    # treated as Javadocs.
    is_java = docs.find('ApiGen') == -1

    # Loop over all the sections of the file that may contain method
    # declarations
    sections = []
    if is_java:
        # Javadoc method declarations are in <pre> sections
        section_pattern = re.compile(r'(?s)<pre>(.*?)</pre>')
        method_pattern = re.compile(r'''(?mx)
            ^([^(\s]+\s+)+?     # Method modifiers and return type
            [a-zA-Z0-9_$]+\s*   # Method name
            \([^),]+,\s*\n      # First method parameter
            (\s*[^),]+,\s*\n)*  # More method parameters
            \s*[^),]+\)         # Final method parameter
            ''')
        throws_pattern = re.compile(r'''(?mx)
            ^\s*throws\s+[^,]+(Throwable|Exception|Error)[^,]*  # First exception
            (,\s*\n\s+[^,]+(Throwable|Exception|Error)[^,]*)*   # More exceptions
            ''')
    else:
        # PHP method declarations are in single-line <code> sections, preceded
        # by '<a class="anchor" href="#_...">#</a>'. Note that the <code>
        # section may contain nested <code> blocks if the method parameter
        # types are links to other classes.
        section_pattern = re.compile(r'''(?x)
            <a\s+class="anchor"\s+href="\#_[^"]+">\#</a>\s*
            <code>((.*?(<code>.*?</code>)?)*)</code>
            ''')
        method_pattern = re.compile(r'''(?x)
            ^[^(]+          # Method name (typically inside a link)
            \(([^),]+)?     # First method parameter
            (,[^),]+)*\)    # More method parameters
            ''')

    for section_match in re.finditer(section_pattern, docs):
        section_start = section_match.start(1)
        section_end = section_match.end(1)
        section_text = section_match.group(1)

        # The section may contain HTML &nbsp; entities in place of some
        # spaces. Replace these with spaces to make the method declaration
        # regexp match, but only keep this change if we end up rewriting the
        # method declaration.
        section_text2 = section_text.replace('&nbsp;', ' ')

        # Test if the section contains a matching method declaration (only
        # expect to find one in each section)
        method_match = re.search(method_pattern, section_text2)
        if method_match:
            method_start = method_match.start()
            method_end = method_match.end()
            method_text = method_match.group()

            # Split the method up into its basic signature (everything up to
            # and including the opening bracket) and the method parameters
            # (excluding the closing bracket)
            bracket_pos = method_text.find('(')
            sig = method_text[:bracket_pos+1]
            params = method_text[bracket_pos+1:-1]

            # Determine the indentation required for the parameters after the
            # first one, based on the raw text length of sig
            raw_sig = re.sub('&.+?;', 'X', sig)
            raw_sig = re.sub('<.*?>', '', raw_sig)
            indent = ' '*len(raw_sig)

            # Extract the parameter types and names, and work out the longest
            # raw text length of a parameter type
            param_types = []
            param_names = []
            max_type_len = 0

            if is_java: param_sep = '\n'
            else: param_sep = ','

            for param in params.split(param_sep):
                param = param.strip()
                if is_java:
                    # Search backwards to find the parameter name
                    split_pos = param.rfind(' ')
                else:
                    # Remove any <code> tags in the parameter declaration so
                    # that the font size is the same throughout
                    param = re.sub('</?code>', '', param)

                    # The parameter declaration is expected to be of the form
                    # '<span>type <var>name</var></span>'. Remove the outer
                    # <span> so that the spaces we add for parameter name
                    # alignment are rendered proportionally
                    if param.startswith('<span>') and\
                       param.endswith('</span>'):
                        param = param[6:-7]

                    # Then search backwards for the parameter name
                    split_pos = param.rfind(' <var>')

                if split_pos != -1:
                    # Have a parameter type and a name
                    param_type = param[:split_pos].strip()
                    param_name = param[split_pos+1:]
                    raw_type = re.sub('&.+?;', 'X', param_type)
                    raw_type = re.sub('<.*?>', '', raw_type)
                    if len(raw_type) > max_type_len:
                        max_type_len = len(raw_type)
                    param_types.append(param_type)
                    param_names.append(param_name)
                else:
                    # Leave this parameter untouched
                    param_types.append('')
                    param_names.append(param)

            # Reconstruct the method text, padding each parameter type to the
            # length of the longest one, and indenting all but the first
            # parameter
            method_text = sig
            for idx, param_type in enumerate(param_types):
                if param_type:
                    raw_type = re.sub('&.+?;', 'X', param_type)
                    raw_type = re.sub('<.*?>', '', raw_type)
                    type_len = len(raw_type)
                    if type_len <= max_type_len:
                        param_type += ' '*(max_type_len - type_len + 1)
                if idx > 0:
                    if not is_java: method_text += ','
                    method_text += '\n' + indent
                method_text += param_type + param_names[idx]
            method_text += ')'

            # Reconstruct the section with this change
            section_text = section_text2[:method_start] +\
                           method_text +\
                           section_text2[method_end:]
            section_text2 = section_text

        if is_java:
            # Test if the section contains a "throws" declaration (only
            # expected to come immediately after a method declaration, which
            # may be on a single line, or multiple lines)
            throws_match = re.search(throws_pattern, section_text2)
            if throws_match:
                throws_start = throws_match.start()
                throws_end = throws_match.end()
                throws_text = throws_match.group()

                # Adjust the indentation of the "throws" to be 4 spaces (and
                # 11 spaces for each subsequent exception, so they line up)
                throws_text = re.sub('(?m)^[ ]*', ' '*11, throws_text)[7:]

                # Reconstruct the section with this change
                section_text = section_text2[:throws_start] +\
                               throws_text +\
                               section_text2[throws_end:]

        # Save the (possibly rewritten) section
        sections.append( (section_start, section_end, section_text) )

    # Join all the pieces back togther
    new_docs = ''
    end = 0
    for (section_start, section_end, section_text) in sections:
        new_docs += docs[end:section_start] + section_text
        end = section_end
    new_docs += docs[end:]

    # If the result is different, rewrite the file
    if new_docs != docs:
        with open(filename, 'w') as f:
            f.write(new_docs)
        print('UPDATED')
    else:
        print('OK')

def process_dir(dirname):
    print('Processing directory "%s" ... ' % dirname)

    for (dirpath, dirnames, filenames) in os.walk(dirname):
        for filename in filenames:
            # Skip class-use and package files and non-HTML files
            if not filename.endswith('.html'): continue
            if dirpath.find('class-use') != -1: continue
            if re.match('package-.*[.]html', filename): continue

            # Process the file
            process_file(os.path.join(dirpath, filename))

    print('DONE')

if __name__ == '__main__':

    if len(sys.argv) > 2 or\
       (len(sys.argv) == 2 and\
        (sys.argv[1] == '-h' or sys.argv[1] == '--help')):
        print('Usage %s [doc_dir]' % os.path.basename(sys.argv[0]))
        sys.exit(0)

    if len(sys.argv) == 2:
        process_dir(sys.argv[1])
    else:
        process_dir(os.getcwd())

